#!/usr/bin/env python

from distutils.core import setup

setup(name='jupyLR',
      version='0.3',
      description='GLR parsing for python with on-line ' +
                  'AST validation and basic error reporting.',
      author='Damien "bl0b" Leroux',
      author_email='damien.leroux@gmail.com',
      url='http://bl0b.github.com/jupyLR/',
      packages=['jupyLR'],
     )
